package dareka.processor;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.HttpIOException;
import dareka.common.Logger;

/**
 * HTTP-message̒ۉB
 *
 */
public class HttpHeader {
    public static final String GET = "GET";
    public static final String POST = "POST";
    public static final String HEAD = "HEAD";
    public static final String OPTIONS = "OPTIONS";

    public static final String ACCEPT_ENCODING = "Accept-Encoding"; // [nl]
    public static final String CONNECTION = "Connection";
    public static final String CONNECTION_CLOSE = "close";
    public static final String CONNECTION_KEEP_ALIVE = "keep-alive";
    public static final String CONTENT_TYPE = "Content-Type";
    public static final String CONTENT_LENGTH = "Content-Length";
    public static final String CONTENT_ENCODING = "Content-Encoding";
    public static final String CONTENT_ENCODING_DEFLATE = "deflate";
    public static final String CONTENT_ENCODING_GZIP = "gzip";
    public static final String IDENTITY = "identity";

    // [nl] twb_
    public static final String DATE = "Date";
    public static final String EXPIRES = "Expires";
    public static final String IF_MODIFIED_SINCE = "If-Modified-Since";
    public static final String IF_UNMODIFIED_SINCE = "If-Unmodified-Since";
    public static final String LAST_MODIFIED = "Last-Modified";

    // [nl] Op[^󂯓n߂̃wb_(NGXgOɏ)
    public static final String X_NICOCACHE_NL = "X-NicoCache_nl";
    public static final String XNL_PROXY_MYSELF = "proxy-myself";

    /**
     * Message Headers.
     *
     * from RFC2616 4.2 Message Headers:
     *
     * <pre>
     *        message-header = field-name &quot;:&quot; [ field-value ]
     *        field-name     = token
     *        field-value    = *( field-content | LWS )
     *        field-content  = &lt;the OCTETs making up the field-value
     *                         and consisting of either *TEXT or combinations
     *                         of token, separators, and quoted-string&gt;
     * </pre>
     *
     *
     * 2.2 Basic Rules:
     * <pre>
     *        OCTET          = &lt;any 8-bit sequence of data&gt;
     *        CHAR           = &lt;any US-ASCII character (octets 0 - 127)&gt;
     *        TEXT           = &lt;any OCTET except CTLs,
     *                         but including LWS&gt;
     *        CTL            = &lt;any US-ASCII control character
     *                         (octets 0 - 31) and DEL (127)&gt;
     *        LWS            = [CRLF] 1*( SP | HT )
     *        token          = 1*&lt;any CHAR except CTLs or separators&gt;
     *        separators     = &quot;(&quot; | &quot;)&quot; | &quot;&lt;&quot; | &quot;&gt;&quot; | &quot;@&quot;
     *                       | &quot;,&quot; | &quot;;&quot; | &quot;:&quot; | &quot;\&quot; | &lt;&quot;&gt;
     *                       | &quot;/&quot; | &quot;[&quot; | &quot;]&quot; | &quot;?&quot; | &quot;=&quot;
     *                       | &quot;{&quot; | &quot;}&quot; | SP | HT
     *        quoted-string  = ( &lt;&quot;&gt; *(qdtext | quoted-pair ) &lt;&quot;&gt; )
     *        qdtext         = &lt;any TEXT except &lt;&quot;&gt;&gt;
     *        quoted-pair    = &quot;\&quot; CHAR
     * </pre>
     */
    private static final Pattern MESSAGE_HEADER_PATTERN =
            Pattern.compile("^([^:]+):\\s*(.*)\r?\n");
    /*
     * GR[fBÖɂă:
     *
     * field-content͓{Ȃǂ̔ASCII蓾BɂǂΉ邩B
     *
     * HTTP̂قƂǂfield-contentASCIIȂ̂
     * ISO-8859-1, MS932, UTF-8vȕR[ĥǂgĂ͓B
     * Aꕔ̃uEU͔ASCII܂߂URL𑗂ĂB
     * (Request-LineReferertB[hȂǁB)
     * RFC3986ᔽȂ̂A͂ĂΉKvB
     * ܂AWwb_Content-DispositionASCII܂ݓB
     *
     * ANicoCache̓ł URLConnection#addRequestProperty(String, String)
     * gĂ邱Ƃ◘֐̂߂field-contentStringƂĈĂB
     * ̂߁Afield-content̃oCgV[PXString̃}bsO
     * ǂ̕GR[fBOɊÂčslKvB
     *
     * ǂ̕GR[fBOĝǂ͊ȒPł͂ȂB
     *
     * PɃuEU-T[oԂ̒ʐM𒆌p̂ł΁A
     * field-contentɂĂ͓Ɉӎɂ̂܂܉΂悢B
     * sɂ͕GR[fBOISO-8859-1g΂悢B
     * ۂ̃oCgV[PXMS932UTF-8̕ꍇ́A
     * JavaStringƂĂ͉ɂȂ邪AISO-8859-1
     * oCgV[PXɖ߂ƌ̃oCgV[PXƓɂȂ̂ŁA
     * 邾Ȃ͋NȂB
     *
     * Afield-content URLConnection#addRequestProperty(String, String)
     * ɓnꍇISO-8859-1Ɩ肪NB
     * ̓Iɂ̓uEŨNGXg󂯂āA
     * URLResourceŃT[oɃNGXg𓊂ꍇB
     * URLResouceA̓IɂHttpURLConnectionAɂ̓
     * sun.net.NetworkClient StringoCgV[PXɕϊۂ
     * file.encodinggB{Windowsł͂MS932ɂȂB
     * oCgV[PXISO-8859-1StringɂāAMS932ŃoCgV[PX
     * ɂƁARƂ͈قȂ̂ɂȂB
     * StringɂۂMS932gƂĂA
     * UTF-8̓{ȂǂMS932̑Ή͈͊ÕoCg܂܂Ăꍇ
     * 񂪎Ă܂(u?vɒu)A͂茳ɂ͖߂ȂȂB
     * ʏ̑΍ƂẮAfile.encodingύX邩A
     * URLConnectionĝ߂邩􂪂ȂA
     * ǂe傫ρB
     *
     * ɖȖƂāARequest-Linefield-contentɈقȂ
     * GR[fBOgĂꍇB
     * Request-LineSJIS or MS932Afield-contentUTF-8ȂǁB
     * ̂߃wb_Ŝ1̕GR[fBOg킯ɂ͂ȂB
     *
     * pł͂ȂANicoCacheNCAgƂĐU镑ꍇ
     * ASCIIgKv͌󖳂̂ŖȂB
     *
     * pł͂ȂANicoCacheoriginT[oƂĐU镑ꍇ
     * URLConnectiongȂ̂ōDɕGR[fBO߂B
     * efield-content͊{IContent-Disposition̂݁B
     * uEUɂĉ߂ł镶R[hقȂ̂ŁA
     * HttpHeader̊OwłKvB
     * efield-contentƂɕR[hقȂ\̂ŁA
     * StringoCgV[PXɕϊۂ̕GR[fBO
     * Ή邽߂ɂ͊efield-contentƂɕGR[fBO
     * ێȂ΂ȂȂBŁAStringoCgV[PXւ
     * ϊISO-8859-1ɌŒ肵Ă܂AsetMessageHeader ̍ۂ
     * ISO-8859-1ŕϊƖ]ރoCgɂȂ鉻StringƂ
     * ݒł悤ɂB
     *
     * _B
     *
     * ̃NXł́AoCgV[PXString̕ϊɂ
     * 񂪎ȂISO-8859-1gB
     * ASCII΍ƂāAGR[fBOw肵
     * field-contentǂݏ郁\bhpӂB
     * wb_oCgV[PXƂĎ擾getBytes()pӂB
     * URLConnection͕ʓrURLConnectiongĂƂŉƂB
     */
    private static final Pattern MULTI_TOKEN_SPLIT =
            Pattern.compile("\\s*,\\s*");
    private static final String ISO_8859_1 = "ISO-8859-1";

    private String startLine = null;
    private HttpMessageHeaderHolder messageHeaders =
            new HttpMessageHeaderHolder();

    /**
     * [nl] RFC2822`̓t/擾B
     * @param time 擾Ώۂ̃G|bN̗ݐώ(~b)
     * @return t/Btime̒lȂnull
     */
    public static String getDateString(long time) {
        if (time < 0) {
            return null;
        }

        DateFormat df = getDateFormatForDateField();
        return df.format(new Date(time));
    }

    /**
     * [nl] t/p[XăG|bN̗ݐώ(~b)ŕԂB
     * @param date t/(RFC2822)
     * @return G|bN̗ݐώ(~b)Bp[XoȂ-1
     */
    public static long parseDateString(String date) {
        if (date != null) {
            DateFormat df = getDateFormatForDateField();
            try {
                return df.parse(date).getTime();
            } catch (ParseException e) {
                // ignore
            }
        }

        return -1L;
    }

    // HTTP-date(RFC2822 3.3. Date and Time Specification)
    private static DateFormat getDateFormatForDateField() {
        // RFC2822 allows to omit seconds, but this method does not.
        // "Z" also accepts obsoleted time zone form such as "JST".
        DateFormat df =
                new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z", Locale.US);
        df.setTimeZone(TimeZone.getTimeZone("GMT"));

        return df;
    }

    public HttpHeader(InputStream source) throws IOException {
        if (source == null) {
            throw new IllegalArgumentException("source must not be null");
        }

        init(source);
    }

    public HttpHeader(String source) throws IOException {
        if (source == null) {
            throw new IllegalArgumentException("source must not be null");
        }

        //init(new ByteArrayInputStream(source.getBytes("UTF-8")));
        init(new ByteArrayInputStream(source.getBytes(ISO_8859_1)));
    }

    // [nl] NGXg1sڈȊOUTF-8ŉ߂ (FxRefererinvalidɂȂ΍j(SP1.05)
    @SuppressWarnings("unused")
    private void initByStringImpl(InputStream source) throws IOException, HttpIOException {
        int ch;
        ByteArrayOutputStream line = new ByteArrayOutputStream(512);

        while ((ch = source.read()) != -1) {
            line.write(ch);

            if (ch != '\n') { // go next read() immediately for performance.
                continue;
            }

            // 1sڈȍ~ōŏ̋s(CRLF)wb_I
            if (startLine != null && line.size() == 2 && line.toString().equals("\r\n")) {
                break;
            }

            if (startLine == null) {
                // IE sends additional CRLF after POST request.
                // see http://support.microsoft.com/kb/823099/
                // see http://httpd.apache.org/docs/1.3/misc/known_client_problems.html#trailing-crlf
                if (line.size() == 2) {
                    line.reset();
                    continue;
                }

                startLine = line.toString("ISO-8859-1");
            } else {
                Matcher m = MESSAGE_HEADER_PATTERN.matcher(line.toString("UTF-8"));
                if (m.find()) {
                    messageHeaders.add(m.group(1), m.group(2));
                } else {
                    Logger.warning("invalid header field: " + line);
                }
            }

            line.reset();
        }

        if (startLine == null || ch == -1) {
            throw new HttpIOException("premature end of header: "
                    + line);
        }
    }

    private void init(InputStream source) throws IOException, HttpIOException {
        int ch;
        ByteArrayOutputStream lineBuf = new ByteArrayOutputStream(512);

        while ((ch = source.read()) != -1) {
            lineBuf.write(ch);

            if (ch != '\n') { // go next read() immediately for performance.
                continue;
            }

            // interpret as ISO-8859-1, because it can preserve
            // original byte sequence in Java string.
            String line = lineBuf.toString(ISO_8859_1);

            if (startLine == null) {
                // IE sends additional CRLF after POST request.
                // see http://support.microsoft.com/kb/823099/
                // see http://httpd.apache.org/docs/1.3/misc/known_client_problems.html#trailing-crlf
                if (line.equals("\r\n")) {
                    lineBuf.reset();
                    continue;
                }

                startLine = line;
            } else {
                if (line.equals("\r\n")) {
                    break;
                }

                Matcher m = MESSAGE_HEADER_PATTERN.matcher(line);
                if (m.find()) {
                    messageHeaders.add(m.group(1), m.group(2));
                } else {
                    Logger.warning("invalid header field: " + lineBuf);
                }
            }

            lineBuf.reset();
        }

        if (startLine == null || ch == -1) {
            String header = createIncompleteHeaderString(lineBuf);
            throw new HttpIOException("premature end of header: " + header);
        }
    }

    private String createIncompleteHeaderString(ByteArrayOutputStream lineBuf)
            throws UnsupportedEncodingException {
        String header = messageHeaders.toString();

        if (startLine != null) {
            header = startLine + header;
        }

        if (lineBuf.size() > 0) {
            header = header + lineBuf.toString(ISO_8859_1);
        }

        return header;
    }

    /**
     * wb_Ŝ𕶎ƂĕԂB
     *
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        StringBuilder result = new StringBuilder(startLine);
        result.append(messageHeaders.toString());
        result.append("\r\n");

        return result.toString();
    }

    /**
     * wb_ŜoCgV[PXƂĕԂB
     */
    public byte[] getBytes() {
        try {
            return toString().getBytes(ISO_8859_1);
        } catch (UnsupportedEncodingException e) {
            // never happen
            throw new IllegalStateException("cannot use " + ISO_8859_1);
        }
    }

    /**
     * Content-Length̒lԂB
     *
     * @return Content-Length̒lB݂Ȃꍇ-1B
     */
    public long getContentLength() {
        String value = getMessageHeader(CONTENT_LENGTH);
        if (value == null) {
            return -1;
        } else {
            return Long.parseLong(value);
        }
    }

    /**
     * Content-Length̒lύXB
     *
     * @param contentLength
     */
    public void setContentLength(long contentLength) {
        messageHeaders.put(CONTENT_LENGTH, String.valueOf(contentLength));
    }

    /**
     * [nl] Last-Modified̒lG|bN̗ݐώ(~b)ŕԂB
     * @return Last-Modified̒lB݂Ȃꍇ-1
     */
    public long getLastModified() {
        String value = messageHeaders.get(LAST_MODIFIED);
        return parseDateString(value);
    }

    /**
     * [nl] Last-Modified̒lݒ肷B
     * @param time ݒ肷G|bN̗ݐώ(~b)B̒lȂݒ肵Ȃ
     */
    public void setLastModified(long time) {
        if (time < 0) {
            return;
        }

        DateFormat df = getDateFormatForDateField();
        messageHeaders.put(LAST_MODIFIED, df.format(new Date(time)));
    }

    /**
     * bZ[Wwb_ԂBkey̑啶͓ꎋB
     * keỹ̕bZ[Wwb_ꍇ͍Ō̂̂ԂB
     *
     * @param key tB[h
     * @return bZ[Wwb_B
     */
    public String getMessageHeader(String key) {
        return messageHeaders.get(key);
    }

    /**
     * bZ[Wwb_w肳ꂽZbgɊÂăfR[hĕԂB
     * key̑啶͓ꎋB
     * keỹ̕bZ[Wwb_ꍇ͍Ō̂̂ԂB
     *
     * @param key
     * @param charsetName ZbgB݂ȂZbgw肵ꍇ
     * {@link HttpHeader#getMessageHeader(String)}ƓB
     * @return bZ[Wwb_B
     */
    public String getMessageHeaderOnCharset(String key, String charsetName) {
        String encodedValue = getMessageHeader(key);
        return decodeString(encodedValue, charsetName);
    }

    /**
     * bZ[Wwb_ŜԂB
     *
     * @return bZ[Wwb_B
     */
    public HttpMessageHeaderHolder getMessageHeaders() {
        return messageHeaders;
    }

    /**
     * bZ[Wwb_ݒ肷Bɂwb_㏑B
     *
    * @param key tB[h
    * @param value tB[hl
     */
    public void setMessageHeader(String key, String value) {
        messageHeaders.put(key, value);
    }

    /**
    * bZ[Wwb_w肳ꂽZbgɊÂ
    * oCgV[PXƂĐݒ肷B
    * ɂwb_㏑B
    *
    * @param key tB[h
    * @param value tB[hl
    * @param charsetName ZbgB݂ȂZbgw肵ꍇ
    * {@link HttpHeader#setMessageHeader(String, String)}ƓB
    */
    public void setMessageHeaderOnCharset(String key, String value,
            String charsetName) {
        String encodedValue = encodeString(value, charsetName);

        setMessageHeader(key, encodedValue);
    }

    /**
     * bZ[Wwb_ǉB
     * @param key
     * @param value
     */
    public void addMessageHeader(String key, String value) {
        messageHeaders.add(key, value);
    }

    /**
     * bZ[Wwb_w肳ꂽZbgɊÂ
     * oCgV[PXƂĒǉB
     *
     * @param key tB[h
     * @param value tB[hl
     * @param charsetName ZbgB݂ȂZbgw肵ꍇ
     * {@link HttpHeader#addMessageHeader(String, String)}ƓB
     */
    public void addMessageHeaderOnCharset(String key, String value,
            String charsetName) {
        String encodedValue = encodeString(value, charsetName);

        addMessageHeader(key, encodedValue);
    }

    public static String encodeString(String normalStr, String charsetName) { // [nl]
        if (charsetName == null) { // [nl]
            return normalStr;
        }

        String encodedStr;
        try {
            encodedStr =
                    new String(normalStr.getBytes(charsetName), ISO_8859_1);
        } catch (UnsupportedEncodingException e) {
            encodedStr = normalStr;
        }
        return encodedStr;
    }

    private static String decodeString(String encodedStr, String charsetName) { // [nl]
        if (charsetName == null) { // [nl]
            return encodedStr;
        }

        String normalStr;
        try {
            normalStr =
                    new String(encodedStr.getBytes(ISO_8859_1), charsetName);
        } catch (UnsupportedEncodingException e) {
            normalStr = encodedStr;
        }

        return normalStr;
    }

    /**
     * bZ[Wwb_폜B
     *
     * @param key
     */
    public void removeMessageHeader(String key) {
        messageHeaders.remove(key);
    }

    protected String getStartLine() {
        return startLine;
    }

    protected void setStartLine(String startLine) {
        this.startLine = startLine;
    }

    /**
     * Hop-by-hop Headers폜BRFC2616Œ`Ă͈̂ȉ:
     *
     * <pre>
     *               - Connection (Ƃɗ񋓂Ă)
     *               - Keep-Alive
     *               - Proxy-Authenticate
     *               - Proxy-Authorization
     *               - TE
     *               - Trailer
     *               - Transfer-Encoding
     *               - Upgrade
     * </pre>
     *
     * Wł͂ȂProxy-ConnectionB
     */
    public void removeHopByHopHeaders() {
        removeConnectionAndRelated();

        removeMessageHeader("Keep-Alive");
        // ad hoc treaing for 407
        //removeMessageHeader("Proxy-Authenticate");
        //removeMessageHeader("Proxy-Authorization");
        removeMessageHeader("TE");
        removeMessageHeader("Trailer");
        removeMessageHeader("Transfer-Encoding");
        removeMessageHeader("Upgrade");
        removeMessageHeader("Proxy-Connection");
    }

    protected void removeConnectionAndRelated() {
        String connection = getMessageHeader(CONNECTION);
        if (connection == null) {
            return;
        }

        String[] tokens = MULTI_TOKEN_SPLIT.split(connection);

        for (String token : tokens) {
            removeMessageHeader(token);
        }

        removeMessageHeader(CONNECTION);
    }

    // [nl] nlł̂ݎgpp[^ݒ(rc2.03)
    private Map<String, String> nlParams = new TreeMap<String, String>(); 
    public String getParameter(String key) {
        return nlParams.get(key);
    }
    public Map<String, String> getParameters() {
    	return nlParams;
    }    
    public void setParameter(String key, String value) {
    	nlParams.put(key, value);
    }
    // [nl] end

// LbṼ_E[hpɃwb_̃GR[hwł悤
//	private String headerEncode = null;
	/**
	 * [nl] NicoCache v0.45 ̃}[WɔÃ\bh͋@\ȂȂ܂B
	 * {@link #setMessageHeaderOnCharset(String, String, String)} g
	 * wb_ʂɕZbgw肵ĂB
	 */
	@Deprecated
	public void setHeaderEncode(String encode) {
//		headerEncode = encode;
		Logger.warning("HttpHeader#setHeaderEncode: obsolete method, do not work.");
	}
	/**
	 * [nl] NicoCache v0.45 ̃}[WɔÃ\bh͋@\ȂȂ܂B
	 * @return  "ISO-8859-1" Ԃ
	 */
	@Deprecated
	public String getHeaderEncode() {
//		return headerEncode;
		Logger.warning("HttpHeader#getHeaderEncode: obsolete method, do not work.");
		return ISO_8859_1;
	}
	
	/**
	 * [nl] LbVwb_tB
	 * @param maxAge LbV̗LbA0ȂLbV
	 */
	public void addCacheControlHeaders(int maxAge) {
		if (maxAge > 0) {
			setMessageHeader("Cache-Control", "max-age=" + maxAge);
			setMessageHeader("Expires", HttpHeader.getDateString(
					System.currentTimeMillis() + (long)maxAge * 1000L));
		} else if (maxAge == 0) {
			setMessageHeader("Cache-Control", "no-cache");
			setMessageHeader("Expires", "Thu, 01 Dec 1994 16:00:00 GMT");
			setMessageHeader("Pragma", "no-cache");
		}
	}
}
